import datetime
import re
import base64

from sqlalchemy import Column, Integer, String, DateTime

from rollbot import as_plugin, RollbotResponse, RollbotFailure, ModelBase, pop_arg, as_group_singleton

from config import get_secret


GUY_STAGES = [
""
,
"""\
😎
"""
,
"""\
😎
👔
"""
,
"""\
😎
👔
⛽
"""
,
"""\
😎
👔
⛽
⚡
"""
,
"""\
👍     😎
          👔
          ⛽
          ⚡
"""
,
"""\
👍     😎
🐛     👔
          ⛽
          ⚡
"""
,
"""\
🖕     😎
🐛💤👔
          ⛽
          ⚡
"""
,
"""\
🖕     😎
🐛💤👔🐛
          ⛽
          ⚡
"""
,
"""\
🖕     😟
🐛💤👔🐛
          ⛽ 👢
          ⚡
"""
,
"""\
🖕     😟
🐛💤👔🐛
          ⛽ 👢
          ⚡👊
"""
,
"""\
🖕     😨
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊
"""
,
"""\
🖕     😨
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊
      🎸
"""
,
"""\
🖕     😰
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊
      🎸
  👢
"""
,
"""\
🖕     😰
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊
      🎸    🌽
  👢
"""
,
"""\
🖕     😰
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊
      🎸    🌽
  👢             👢
"""
,
"""\
🖕     😫
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊=D
      🎸    🌽
  👢             👢
"""
,
"""\
🖕     😵
🐛💤👔🐛
          ⛽  👢
          ⚡8=👊=D💦
      🎸    🌽
  👢             👢
"""
,
]

@as_group_singleton
class HangguyGameState:
    target: str = None
    state: str = None
    bad_guesses: str = ""


@as_group_singleton
class HangguyGuyState:
    state: int = 0
    games: int = 0


class DeadGuy(ModelBase):
    """
    Tracks "dead" and "retired" guys for record-keeping purposes.
    If state >= len(GUY_STAGES) - 1, the guy was dead, otherwise it was retired.
    """
    __tablename__ = "hangguy_dead_guy"
    guy_id = Column(Integer, primary_key=True, autoincrement=True)
    group_id = Column(String)
    state = Column(Integer)
    games = Column(Integer)
    timestamp = Column(DateTime)

    @staticmethod
    def from_guy(guy):
        return DeadGuy(
            group_id=guy.group_id,
            state=guy.state,
            games=guy.games,
            timestamp=datetime.datetime.now()
        )


def init_game(game_state, phrase):
    if phrase is None:
        new_phrase = "TEST PHRASE" # TODO this needs to actually pull from something (and sanitize it) eventually
    else:
        new_phrase = base64.b64decode(phrase).decode("utf-8").strip().upper()

    game_state.state = re.sub("[A-Z]", "_", new_phrase)
    game_state.target = new_phrase
    game_state.bad_guesses = ""


def end_game(game_state):
    game_state.target = None


def game_is_active(game_state):
    return game_state.target is not None


def game_is_finished(game_state):
    return game_state.target == game_state.state


def guess_letter(game_state, letter):
    """
    Guess a letter in a given game state. Assumes letter is already
    a single, uppercased, alpha character.

    Returns number of appearances of the letter, or -1 if it was
    already guessed.
    """

    if letter in game_state.bad_guesses or letter in game_state.state:
        return -1

    find = [i for i, x in enumerate(game_state.target) if x == letter]
    if len(find) == 0:
        game_state.bad_guesses = game_state.bad_guesses + letter
        return 0

    game_state.state = "".join(letter if i in find else s for i, s in enumerate(game_state.state))
    return len(find)


def end_guy(db, guy_state):
    db.add(DeadGuy.from_guy(guy_state))
    guy_state.state = 0
    guy_state.games = 0


def guy_is_dead(guy_state):
    return guy_state.state >= (len(GUY_STAGES) - 1)


def survival_msg(guy_state):
    return f"survived {guy_state.games} game{'' if guy_state.games == 1 else 's'}"


def render_guy(guy_state):
    if 0 <= guy_state.state < len(GUY_STAGES):
        return GUY_STAGES[guy_state.state]
    elif guy_state.state >= len(GUY_STAGES):
        return GUY_STAGES[-1]
    return "INVALID GUY STATE " + str(guy_state.state)


def render_game(game_state, guy_state):
    return render_guy(guy_state) + "\n" + \
        " ".join(game_state.state) + "\n" + \
        "Bad Guesses: " + ", ".join(game_state.bad_guesses)


def handle_guess(db, guess, game_state, guy_state):
    prefix = ""
    if len(guess) == 1:
        if guess.isalpha():
            res = guess_letter(game_state, guess)
            if res == -1:
                prefix += f"You've already guessed '{guess}'!"
            elif res == 0:
                prefix += "Bad guess!"
                guy_state.state = guy_state.state + 1
            else:
                prefix += f"Great! '{guess}' appears {res} time{'' if res == 1 else 's'}!"
        else:
            prefix += "You should try sticking to letters!"
    elif game_state.target.split() == guess.split():
        prefix = "You've guessed the full phrase!"
        game_state.state = game_state.target
    else:
        prefix = f"{guess} is not the phrase!"
        guy_state.state = guy_state.state + 2

    txt = prefix + "\n" + render_game(game_state, guy_state)

    if game_is_finished(game_state):
        txt += "\nThe game is over, and the guy lives on!"
        end_game(game_state)
        guy_state.games = guy_state.games + 1
        txt += f"\nThis guy has now {survival_msg(guy_state)}"

    if guy_is_dead(guy_state):
        txt += "\noh god oh fuck, the guy is dead!"
        txt += f"\nYou failed to guess {game_state.target}"
        end_game(game_state)
        end_guy(db, guy_state)
        txt += "\nThe guy has been buried."

    return txt


def subc_retire(db, guy_state):
    txt = f"Retiring the previous guy, who {survival_msg(guy_state)}\n"
    txt += render_guy(guy_state)
    end_guy(db, guy_state)
    return txt


def subc_alert(bot, msg):
    bot.manually_post_message(f"{msg.name} wants you to check out the hangguy chat!", get_secret("hangguy.alert_chat"))
    return None # do not respond in the current chat


@as_plugin
def hangguy(bot, db, msg, subc, game_data: HangguyGameState, guy_data: HangguyGuyState):
    if msg.raw_args is None:
        return RollbotFailure.MISSING_SUBCOMMAND.with_reason("Must provide subcommand or guess")

    if game_is_active(game_data):
        if subc is None:
            return handle_guess(db, msg.raw_args.upper(), game_data, guy_data)   
        elif subc.command == "cancel":
            end_game(game_data)
            return "The game has been cancelled. The guy has not been reset."
        elif subc.command == "alert":
            return subc_alert(bot, msg)
        else:
            return RollbotFailure.INVALID_SUBCOMMAND.with_reason(
                "The only in-game subcommands are !cancel to end the game and !alert to alert the main chat."
            )

    if subc is None:
        return RollbotFailure.MISSING_SUBCOMMAND.with_reason(
            "You are not currently in a game, and so you must use one of the subcommands: !view, !retire, !start, !alert"
        )
    elif subc.command == "view":
        return f"The current guy has {survival_msg(guy_data)}\n" \
            + render_guy(guy_data)
    elif subc.command == "retire":
        return subc_retire(db, guy_data)
    elif subc.command == "start":
        init_game(game_data, subc.raw_args)
        return render_game(game_data, guy_data)
    elif subc.command == "alert":
        return subc_alert(bot, msg)
    else:
        return RollbotFailure.INVALID_SUBCOMMAND.with_reason(
            "You must use one of the subcommands: !view, !retire, !start, !alert"
        )
